!
! Module: cal_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: A few routines to obtain basic information from the cal.
! 
! Comments: TODO:
! 
! $Id: cal_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module cal_aux

  use dal
  use cal
  use oal
  use errorHandling

  use constants
  use coordinate	! for getInstrumentAttitude()

  implicit none 

  private 

  public getCalInfo

  interface getCalInfo
    module procedure getCalInfoName, getCalInfoSet
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine getCalInfoName(setName, instrumentId, expStartTimeStamp&
  , scAttitude, instrumAttitude)
  !---------------------------------------------------------------------
  ! A routine to extract (directly or indirectly) the instrumentId,
  ! expStartTimeStamp, scAttitude and instrumAttitude from the header
  ! of the specified data set.
  !
  ! NOTE! This routine also sets the instrument state of the cal.

    implicit none

    character(*),                 intent(in)  :: setName
    integer(int32),               intent(out) :: instrumentId
    real(double),                 intent(out) :: expStartTimeStamp
    type(SpacecraftAttitudeType), intent(out) :: scAttitude, instrumAttitude

    type(DataSetT) :: set

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    call getCalInfoSet(set, instrumentId, expStartTimeStamp, scAttitude&
      , instrumAttitude)
    call release(set)

  end subroutine getCalInfoName

  !---------------------------------------------------------------------
  subroutine getCalInfoSet(set, instrumentId, expStartTimeStamp, scAttitude&
  , instrumAttitude)
  !---------------------------------------------------------------------
  ! A routine to extract (directly or indirectly) the instrumentId,
  ! expStartTimeStamp, scAttitude and instrumAttitude from the header
  ! of the specified data set.
  !
  ! NOTE! This routine also sets the instrument state of the cal.
  ! CONSTRAINTS: The instrument has to be EPIC; otherwise raises an error.

    implicit none

    type(DataSetT),               intent(in)  :: set
    integer(int32),               intent(out) :: instrumentId
    real(double),                 intent(out) :: expStartTimeStamp
    type(SpacecraftAttitudeType), intent(out) :: scAttitude, instrumAttitude

    character(20) :: instrumentName, dateTimeStr
    real(double) ::&
      pntRaDeg,&
      pntDecDeg,&
      pntAposDeg

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Noisy, 'ssclib: retrieve instrum, dateobs, attitude.')

    instrumentName = stringAttribute(set, "INSTRUME")
    instrumentId = instId(instrumentName) ! in caloalutils/src/caloaldefs.f90

    if (instrumentId .ne. EMOS1&
    .and. instrumentId .ne. EMOS2&
    .and. instrumentId .ne. EPN) then
call error('badInstrum', 'Instrument '//trim(instrumentName)//' not supported.')
    end if

    pntRaDeg   = real64Attribute(set, "RA_PNT")
    pntDecDeg  = real64Attribute(set, "DEC_PNT")
    pntAposDeg = real64Attribute(set, "PA_PNT")

    scAttitude%ra   = pntRaDeg   * PiOn180
    scAttitude%dec  = pntDecDeg  * PiOn180
    scAttitude%apos = pntAposDeg * PiOn180

!******* It actually SHOULD be the INSTRUMENT attitude, but was still s/c, as of 1 Feb 2002, in both MOS and PN event lists. When the change comes along we will need to read RA_SCX, _SCY etc kwds to get the s/c attitude. Handy to have a CAL call to convert SCX/Y/Z to scAttitude... Uwe??

    dateTimeStr = stringAttribute(set, "DATE-OBS")
    expStartTimeStamp = toTimeTag(dateTimeStr) ! in caloalutils/caloaldefs.f90

    call CAL_setState(instrument=instrumentId, randomize=.false.)
    instrumAttitude = getInstrumentAttitude(scAttitude, expStartTimeStamp)	! ssclib/coordinate

  end subroutine getCalInfoSet

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module cal_aux
