!+
! Module: sort_mod
!
! Authors:  Clive Page - SSC/LUX
!          
! Description: Non-recursive stack versions of quicksort from N.Wirth's Pascal
!  book, "Algorithms + data structures = programs".
!
! Converted to Fortran90 by Alan Miller, CSIRO.
!
! Module constructed by Clive Page, Leicester University.
!
! Some changes to format, + quickSortStatic added, by I Stewart 13 May 2003.
!
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module sort_mod

  use types		! For single, double, int32 etc type definitions.

  use ssc_misc	! for swap(), isNaN().


  implicit none

  private 

  public quickSort
  public quickSortStatic

  interface quickSort
     module procedure quickSortStr,     quickSortStrWithIndex
     module procedure quickSortDouble,  quickSortDoubleWithIndex
     module procedure quickSortInt8
     module procedure quickSortInt16
     module procedure quickSortInt32,   quickSortInt32WithIndex
     module procedure quickSortSingle,  quickSortSingleWithIndex
  end interface quickSort

  interface quickSortStatic
     module procedure quickSortStaticSingle, quickSortStaticDouble
  end interface quickSortStatic

contains

  !---------------------------------------------------------------------
  subroutine quickSortStaticSingle(values, indices)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: values(:)
    integer,      intent(out) :: indices(size(values))

    real(single), allocatable :: tempValues(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(tempValues(size(values)))

    tempValues = values
    call quickSortSingleWithIndex(tempValues, indices)

  end subroutine quickSortStaticSingle

  !---------------------------------------------------------------------
  subroutine quickSortStaticDouble(values, indices)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: values(:)
    integer,      intent(out) :: indices(size(values))

    real(double), allocatable :: tempValues(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(tempValues(size(values)))

    tempValues = values
    call quickSortDoubleWithIndex(tempValues, indices)

  end subroutine quickSortStaticDouble

  !---------------------------------------------------------------------
  subroutine quickSortStr(array)
  !---------------------------------------------------------------------

    character(*), intent(INOUT) :: array(:)

    !     Local Variables

    character(LEN=len(array)) :: x, temp
    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    !-------------------------------------------------------------------
    ! Start

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                array(i) = array(j)
                array(j) = temp
!!!call swap(array(i), array(j))
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do

  end subroutine quickSortStr

  !---------------------------------------------------------------------
  subroutine quickSortInt8(int8Array)
  !---------------------------------------------------------------------

    integer(int8), intent(INOUT) :: int8Array(:)

    integer(int32) :: int32Array(size(int8Array))

    !-------------------------------------------------------------------
    ! Start

    int32Array = int8Array
    call quickSortInt32(int32Array)
    int8Array = int32Array

  end subroutine quickSortInt8

  !---------------------------------------------------------------------
  subroutine quickSortInt16(int16Array)
  !---------------------------------------------------------------------

    integer(int16), intent(INOUT) :: int16Array(:)

    integer(int32) :: int32Array(size(int16Array))

    !-------------------------------------------------------------------
    ! Start

    int32Array = int16Array
    call quickSortInt32(int32Array)
    int16Array = int32Array

  end subroutine quickSortInt16

  !---------------------------------------------------------------------
  subroutine quickSortInt32(array)
  !---------------------------------------------------------------------

    integer(int32), intent(INOUT) :: array(:)

    !     Local Variables

    integer(int32) :: x
!temp, 
    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    !-------------------------------------------------------------------
    ! Start

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
!                temp    = array(i)
!                array(i) = array(j)
!                array(j) = temp
call swap(array(i), array(j))
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do

  end subroutine quickSortInt32

  !---------------------------------------------------------------------
  subroutine quickSortSingle(array, status)
  !---------------------------------------------------------------------

    real(single), intent(INOUT)         :: array(:)
    integer,      intent(out), optional :: status

    !     Local Variables

    real(single) :: x
!temp, 
    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    !-------------------------------------------------------------------
    ! Start

    if (present(status)) status = 0

    if (any(isNaN(array))) then
      if (present(status)) status = -1
  return
    end if

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
!                temp    = array(i)
!                array(i) = array(j)
!                array(j) = temp
call swap(array(i), array(j))
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do

  end subroutine quickSortSingle

  !---------------------------------------------------------------------
  subroutine quickSortDouble(array, status)
  !---------------------------------------------------------------------

    real(double), intent(INOUT)         :: array(:)
    integer,      intent(out), optional :: status

    !     Local Variables

    real(double) :: x
!temp, 
    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    !-------------------------------------------------------------------
    ! Start

    if (present(status)) status = 0

    if (any(isNaN(array))) then
      if (present(status)) status = -1
  return
    end if

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
!                temp    = array(i)
!                array(i) = array(j)
!                array(j) = temp
call swap(array(i), array(j))
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do

  end subroutine quickSortDouble

  !---------------------------------------------------------------------
  subroutine quickSortStrWithIndex(array, order)
  !---------------------------------------------------------------------

    character(*), intent(INOUT) :: array(:)
    integer,      intent(OUT)   :: order(:)  ! returns original position

    !     Local Variables

    character(LEN=len(array)) :: x, temp
    integer  :: i, j, k, l, r, s, stackl(15), stackr(15)
!, ww

    !-------------------------------------------------------------------
    ! Start

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
!                ww   = order(i)
                array(i) = array(j)
!                order(i) = order(j)
                array(j) = temp
!                order(j) = ww
!call swap(array(i), array(j))
call swap(order(i), order(j))
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do

  end subroutine quickSortStrWithIndex

  !---------------------------------------------------------------------
  subroutine quickSortInt32WithIndex(array, order)
  !---------------------------------------------------------------------

    integer(int32), intent(INOUT) :: array(:)
    integer,        intent(OUT)   :: order(:)  ! returns original position

    !     Local Variables

    integer(int32)  :: x
!temp, 
    integer  :: i, j, k, l, r, s, stackl(15), stackr(15)
!, ww

    !-------------------------------------------------------------------
    ! Start

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
!                temp    = array(i)
!                ww   = order(i)
!                array(i) = array(j)
!                order(i) = order(j)
!                array(j) = temp
!                order(j) = ww
call swap(array(i), array(j))
call swap(order(i), order(j))
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do

  end subroutine quickSortInt32WithIndex

  !---------------------------------------------------------------------
  subroutine quickSortSingleWithIndex(array, order, status)
  !---------------------------------------------------------------------

    real(single), intent(INOUT)         :: array(:)
    integer,      intent(OUT)           :: order(:)  ! returns original position
    integer,      intent(out), optional :: status

    !     Local Variables

    real(single) :: x
!temp, 
    integer  :: i, j, k, l, r, s, stackl(15), stackr(15)
!, ww

    !-------------------------------------------------------------------
    ! Start

    if (present(status)) status = 0

    if (any(isNaN(array))) then
      if (present(status)) status = -1
  return
    end if

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
!                temp    = array(i)
!                ww   = order(i)
!                array(i) = array(j)
!                order(i) = order(j)
!                array(j) = temp
!                order(j) = ww
call swap(array(i), array(j))
call swap(order(i), order(j))
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do

  end subroutine quickSortSingleWithIndex

  !---------------------------------------------------------------------
  subroutine quickSortDoubleWithIndex(array, order, status)
  !---------------------------------------------------------------------

    real(double), intent(INOUT)         :: array(:)
    integer,      intent(OUT)           :: order(:)  ! returns original position
    integer,      intent(out), optional :: status

    !     Local Variables

    real(double) :: x
!temp, 
    integer  :: i, j, k, l, r, s, stackl(15), stackr(15)
!, ww

    !-------------------------------------------------------------------
    ! Start

    if (present(status)) status = 0

    if (any(isNaN(array))) then
      if (present(status)) status = -1
  return
    end if

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
             if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
             if( x >= array(j)) exit
                j = j - 1
             end do

          if (i > j) exit

             ! Otherwise, swap positions i & j

!             temp    = array(i)
!             ww   = order(i)
!             array(i) = array(j)
!             order(i) = order(j)
!             array(j) = temp
!             order(j) = ww
call swap(array(i), array(j))
call swap(order(i), order(j))
             i = i + 1
             j = j - 1
          if (i > j) exit
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if

       if(l >= r) exit
       end do
    if (s == 0) exit
    end do

  end subroutine quickSortDoubleWithIndex

  !---------------------------------------------------------------------
  subroutine quickSortDoubleWithIndex_old(array, order) ! 20041109
  !---------------------------------------------------------------------
  ! New routine includes swap, tests for NaNs, and is slightly rearranged.

    real(double), intent(INOUT)    :: array(:)
    integer, intent(OUT) :: order(:)  ! returns original position
    real(double) :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(15), stackr(15), ww

    !-------------------------------------------------------------------
    ! Start

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
             if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
             if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                ww   = order(i)
                array(i) = array(j)
                order(i) = order(j)
                array(j) = temp
                order(j) = ww
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
          exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
!    return

  end subroutine quickSortDoubleWithIndex_old ! 20041109

end module sort_mod
