program testdetectionstats
  use dal
  use errorhandling
  use test_utils	! for isNearlyEqual()
  use detection_stats

  implicit none

  call test_integratedChi2Prob

contains

  subroutine test_integratedChi2Prob
    real(single) :: chi2 = 2.706
    integer      :: idegFree = 1
    real(single) :: fdegFree = 1
    real(single) :: probability	! to receive
    integer      :: status	! to receive

    !! Following example values from e.g. Table C.5 in Bevington and Robinson.

    call integratedChi2Prob(chi2, idegFree, probability, status)	! integratedChi2ProbInteger
                          ! 2.706,  1
    if (status .ne. 0) then
      write(errstr, *) "Status = ", status
      call error("Chi2ProbIntegerStatus00")
    else if ( .not. isNearlyEqual(0.100, probability, 3) ) then
      write(errstr, *) "Probability(Integer1)(=0.10) = ", probability
      call error("Chi2ProbIntegerProb00")
    end if

    call integratedChi2Prob(2.303*2.0, 2, probability, status)
    if (status .ne. 0) then
      write(errstr, *) "Status = ", status
      call error("Chi2ProbIntegerStatus05")
    else if ( .not. isNearlyEqual(0.100, probability, 3) ) then
      write(errstr, *) "Probability(Integer2)(=0.10) = ", probability
      call error("Chi2ProbIntegerProb05")
    end if


    call integratedChi2Prob(chi2, fdegFree, probability, status)	! integratedChi2ProbSingle
                          ! 2.706,  1
    if (status .ne. 0) then
      write(errstr, *) "Status = ", status
      call error("Chi2ProbSingleStatus")
    else if ( .not. isNearlyEqual(0.100, probability, 3) ) then
      write(errstr, *) "Probability(Single1)(=0.10) = ", probability
      call error("Chi2ProbSingleProb")
    end if
  end subroutine test_integratedChi2Prob


end program testdetectionstats
