!
! Program: testdss_aux02
!
! Authors:  Masaaki Sakano - SSC/LUX
!          
! Description: Test dss_aux module, Part 2.
! 
! Comments:
!
!-
program testdss_aux02
  ! use cal	! for convertPiToEeV (and CAL_setState)
  use caltypes	! for IntervalT
  use dal
  use dss	! for RangeT and BoundT
  use errorhandling
  use dal_aux
  ! use test_utils	! for isNearlyEqual()
  use dss_aux
  ! use dss_ranges_aux	! for copyRanges
  ! use exposure	! for convertPiToEeV

  implicit none

  character(256), parameter :: INFILE = 'outfiles/P0111360101M1S002MIEVLI0000.FIT'
  integer, parameter :: &	! See dsslib/src/dsscore/dssinter.f90 for the definition of 80 and 255
       dssFiltCharLength = 80 &
     , dssFilterBlockCharLength = 255
  integer(int32), parameter :: maxRepM = 1
  !integer(int32), parameter :: maxRepM = 100	! for (memory) stress test.
  ! It takes 110sec ish with aral for maxRep=100 (in 2011).
  integer(int32) :: iM !, nM

  !-------------------------------------------------------------------
  ! Start:

  do iM=1,maxRepM
    if (maxRepM > 1) print *,'*Repetition No.',iM
    call test_allFiltersPassValue
    call test_getNum1stDssCompWithValue
    call test_readGtisFromDss
  end do

contains

  !---------------------------------------------------------------------
  subroutine test_allFiltersPassValue
  !---------------------------------------------------------------------

    implicit none

    character(32) :: errHead 
    type(DataSetT) :: inSet
    type(BlockT)   :: eventBlock
    type(DssT)    :: dssPointer
    type(DScompT) :: dssComp
    ! integer :: compNum
    logical :: isTrue

    !-------------------------------------------------------------------
    ! Start:

    inSet = dataSet(trim(INFILE), READ)
    eventBlock = block(inSet, 'EVENTS')
    if (.not. hasDss(eventBlock)) call error('bug0010',trim(INFILE)//' may not exist.')
    dssPointer = dataSubspace(eventBlock)

    ! do compNum = 0, dssComponents(dssPointer) - 1
    dssComp = dssComponent(dssPointer, 3)	! 3 for CCD=05
    ! call dumpDss(dssComp)

    !----------------------------------------------------------
    errHead = "AllFiltersPassValue0010a"
    print *,"=== "//trim(errHead)//": Standard"

    isTrue = allFiltersPassValue(dssComp, 'CCDNR', 5.0)

    if (.not. isTrue) then
      write(errStr, '("isTrue =",l2)') isTrue
      call error('ERRallFiltersPass0010a')
    end if

    !----------------------------------------------------------
    errHead = "AllFiltersPassValue0020a"
    print *,"=== "//trim(errHead)//": Standard with filterType"

    isTrue = allFiltersPassValue(dssComp, 'CCDNR', 5.0 &
       , filterType=RANGEFILTERT)

    if (.not. isTrue) then
      write(errStr, '("isTrue =",l2)') isTrue
      call error('ERRallFiltersPass0020a')
    end if

    !----------------------------------------------------------
    errHead = "AllFiltersPassValue0030a"
    print *,"=== "//trim(errHead)//": Standard with a WRONG filterType"

    isTrue = allFiltersPassValue(dssComp, 'CCDNR', 5.0 &
       , filterType=GTIFILTERT)
    if (.not. isTrue) then
      write(errStr, '("isTrue =",l2)') isTrue
      call error('ERRallFiltersPass0030a')
    end if

    !----------------------------------------------------------
    errHead = "AllFiltersPassValue0040a"
    print *,"=== "//trim(errHead)//": Standard with a WRONG filterType with trueIfNone=F"

    isTrue = allFiltersPassValue(dssComp, 'CCDNR', 5.0 &
       , filterType=GTIFILTERT, trueIfNone=.false.)
    if (isTrue) then
      write(errStr, '("isTrue =",l2)') isTrue
      call error('ERRallFiltersPass0040a')
    end if

    call dssRelease(dssPointer)
    call release(inSet) 

  end subroutine test_allFiltersPassValue


  !---------------------------------------------------------------------
  subroutine test_getNum1stDssCompWithValue
  !---------------------------------------------------------------------

    implicit none

    character(32) :: errHead 
    type(DataSetT) :: inSet
    type(BlockT)   :: eventBlock
    type(DssT)     :: dssPointer
    integer :: firstCompNum
    integer :: status

    character(dssFilterBlockCharLength) :: gtiBlockName

    !-------------------------------------------------------------------
    ! Start:

    inSet = dataSet(trim(INFILE), READ)
    eventBlock = block(inSet, 'EVENTS')
    if (.not. hasDss(eventBlock)) call error('bug0010',trim(INFILE)//' may not exist.')
    dssPointer = dataSubspace(eventBlock)


    !----------------------------------------------------------
    errHead = "GetNum1stDssCompWithValue0010"
    print *,"=== "//trim(errHead)//": Standard"

    call getNum1stDssCompWithValue(dssPointer, 'CCDNR', 5.0, firstCompNum &
       , status=status)

    if (status .ne. 0) then
      write(errStr, '("status =",i5)') status
      call error('ERRgetNum1st0010a')
    end if
    if (firstCompNum .ne. 3) then
      write(errStr, '("firstCompNum =",i5)') firstCompNum
      call error('ERRgetNum1st0010b')
    end if

    call getGtiBlockName(dssPointer, firstCompNum, gtiBlockName)
    if (trim(gtiBlockName) .ne. trim(INFILE)//':STDGTI05') then
      write(errStr, '("gtiBlockName = ",a)') trim(gtiBlockName)
      call error('ERRgetNum1st0010c')
    end if


    !----------------------------------------------------------
    errHead = "GetNum1stDssCompWithValue0020"
    print *,"=== "//trim(errHead)//": Standard with filterType"

    call getNum1stDssCompWithValue(dssPointer, 'CCDNR', 5.0, firstCompNum &
       , filterType=RANGEFILTERT, status=status)

    if (status .ne. 0) then
      write(errStr, '("status =",i5)') status
      call error('ERRgetNum1st0020a')
    end if
    if (firstCompNum .ne. 3) then
      write(errStr, '("firstCompNum =",i5)') firstCompNum
      call error('ERRgetNum1st0020b')
    end if

    call getGtiBlockName(dssPointer, firstCompNum, gtiBlockName)
    if (trim(gtiBlockName) .ne. trim(INFILE)//':STDGTI05') then
      write(errStr, '("gtiBlockName = ",a)') trim(gtiBlockName)
      call error('ERRgetNum1st0020c')
    end if

    !----------------------------------------------------------
    errHead = "GetNum1stDssCompWithValue0030"
    print *,"=== "//trim(errHead)//": Standard with a WRONG filterType with trueIfNone=T."

    call getNum1stDssCompWithValue(dssPointer, 'CCDNR', 5.0, firstCompNum &
       , filterType=GTIFILTERT, trueIfNone=.true., status=status)

    if (status .ne. 0) then
      write(errStr, '("status =",i5)') status
      call error('ERRgetNum1st0030a')
    end if
    if (firstCompNum .ne. 0) then	! No GTIFILTERT is found in the first DSS-component, hence it is true.
      write(errStr, '("firstCompNum =",i5)') firstCompNum
      call error('ERRgetNum1st0030b')
    end if

    !----------------------------------------------------------
    errHead = "GetNum1stDssCompWithValue0040"
    print *,"=== "//trim(errHead)//": Standard with a WRONG filterType with trueIfNone=F."

    call getNum1stDssCompWithValue(dssPointer, 'CCDNR', 5.0, firstCompNum &
       , filterType=GTIFILTERT, trueIfNone=.false., status=status)

    if (status > -1) then
      write(errStr, '("status =",i5)') status
      call error('ERRgetNum1st0030a')
    end if
    if (firstCompNum > -1) then
      write(errStr, '("firstCompNum =",i5)') firstCompNum
      call error('ERRgetNum1st0030b')
    end if

    !-----
    ! The following must raise
    !   error (noDssCompThisValue)
    ! But it can not be implemented in the test harness.
    ! Therefore best uncomment it and run it by yourself in order to test it.
    !-----
    ! call getNum1stDssCompWithValue(dssPointer, 'CCDNR', 5.0, firstCompNum &
    !    , filterType=GTIFILTERT, trueIfNone=.false.)

    !----------------------------------------------------------
    ! Ending

    call dssRelease(dssPointer)
    call release(inSet) 

  end subroutine test_getNum1stDssCompWithValue


  !---------------------------------------------------------------------
  subroutine test_readGtisFromDss
  !---------------------------------------------------------------------

    implicit none

    character(32) :: errHead 
    type(DataSetT) :: inSet
    type(BlockT)   :: eventBlock
    type(TableT)   :: tabGtiCcd05
    type(DssT)     :: dssPointer
    type(IntervalT), pointer :: gti(:)	! IntervalT defined in caltypes
    real(double),    pointer :: startColGti(:), stopColGti(:)
    integer :: sizeGti

    !-------------------------------------------------------------------
    ! Start:

    ! print *, 'Test test_readGtisFromDss'
    inSet = dataSet(trim(INFILE), READ)
    eventBlock = block(inSet, 'EVENTS')
    if (.not. hasDss(eventBlock)) call error('bug0010',trim(INFILE)//' may not exist.')
    dssPointer = dataSubspace(eventBlock)


    !----------------------------------------------------------
    errHead = "ReadGtisFromDss0010"
    print *,"=== "//trim(errHead)//": Standard (Block)"

    ! The following can be done with  subroutine readGtisSet()  in dal_aux
    tabGtiCcd05 = table(inSet, 'STDGTI05')		! CCD Chip No.5
    call readColDataToPtr(tabGtiCcd05, 'START', startColGti)	! defined in dal_aux
    call readColDataToPtr(tabGtiCcd05, 'STOP',  stopColGti)	! defined in dal_aux

    ! Tested subroutine.
    call readGtisFromDss(inSet, eventBlock, 5, gti)	! CCD Chip No.5
    sizeGti = size(gti)

    if (sizeGti .ne. size(startColGti) .or. &
        sizeGti .ne. size(stopColGti) ) then
      write(errStr, '("Size gti(:)= ",i4," differs from Column ",i4)') size(gti), size(startColGti)
      call error('Err'//trim(errHead))
    else if (gti(1)%lower .ne. startColGti(0)) then
      print *, 'gti(:)%lower   = ', gti%lower
      print *, 'startColGti(:) = ', startColGti
      write(errStr, '("gti(1)%lower =",es12.6)') gti(1)%lower
      call error('Err'//trim(errHead))
    else if (gti(1)%upper .ne. stopColGti(0)) then
      print *, 'gti(:)%upper  = ', gti%upper
      print *, 'stopColGti(:) = ', stopColGti
      write(errStr, '("gti(1)%upper =",es12.6)') gti(1)%upper
      call error('Err'//trim(errHead))
    else if (gti(sizeGti)%lower .ne. startColGti(sizeGti-1)) then
      print *, 'gti(:)%lower   = ', gti%lower
      print *, 'startColGti(:) = ', startColGti
      write(errStr, '("gti(",i2,")%lower =",es12.6)') sizeGti, gti(sizeGti)%lower
      call error('Err'//trim(errHead))
    else if (gti(sizeGti)%upper .ne. stopColGti(sizeGti-1)) then
      print *, 'gti(:)%upper  = ', gti%upper
      print *, 'stopColGti(:) = ', stopColGti
      write(errStr, '("gti(",i2,")%upper =",es12.6)') sizeGti, gti(sizeGti)%upper
      call error('Err'//trim(errHead))
    end if

    deallocate(gti)


    !----------------------------------------------------------
    errHead = "ReadGtisFromDss0020"
    print *,"=== "//trim(errHead)//": Standard (Pointer)"

    ! Tested subroutine.
    call readGtisFromDss(inSet, dssPointer, 5, gti)	! CCD Chip No.5
    sizeGti = size(gti)

    !---- The following is the copy of the Test ReadGtisFromDss0010
    if (sizeGti .ne. size(startColGti) .or. &
        sizeGti .ne. size(stopColGti) ) then
      write(errStr, '("Size gti(:)= ",i4," differs from Column ",i4)') size(gti), size(startColGti)
      call error('Err'//trim(errHead))
    else if (gti(1)%lower .ne. startColGti(0)) then
      print *, 'gti(:)%lower   = ', gti%lower
      print *, 'startColGti(:) = ', startColGti
      write(errStr, '("gti(1)%lower =",es12.6)') gti(1)%lower
      call error('Err'//trim(errHead))
    else if (gti(1)%upper .ne. stopColGti(0)) then
      print *, 'gti(:)%upper  = ', gti%upper
      print *, 'stopColGti(:) = ', stopColGti
      write(errStr, '("gti(1)%upper =",es12.6)') gti(1)%upper
      call error('Err'//trim(errHead))
    else if (gti(sizeGti)%lower .ne. startColGti(sizeGti-1)) then
      print *, 'gti(:)%lower   = ', gti%lower
      print *, 'startColGti(:) = ', startColGti
      write(errStr, '("gti(",i2,")%lower =",es12.6)') sizeGti, gti(sizeGti)%lower
      call error('Err'//trim(errHead))
    else if (gti(sizeGti)%upper .ne. stopColGti(sizeGti-1)) then
      print *, 'gti(:)%upper  = ', gti%upper
      print *, 'stopColGti(:) = ', stopColGti
      write(errStr, '("gti(",i2,")%upper =",es12.6)') sizeGti, gti(sizeGti)%upper
      call error('Err'//trim(errHead))
    end if

    deallocate(gti)

    !----------------------------------------------------------
    ! Ending

    deallocate(startColGti, stopColGti)
    call dssRelease(dssPointer)
    call release(inSet) 

  end subroutine test_readGtisFromDss


  !---------------------------------------------------------------------
  subroutine getGtiBlockName(dssPointer, compNum, gtiBlockName)
  !---------------------------------------------------------------------
  ! Get GTI_BlockName for a given DSSComp number.

    implicit none

    type(DssT), intent(in) :: dssPointer
    integer,    intent(in) :: compNum
    character(dssFilterBlockCharLength), intent(out) :: gtiBlockName

    type(DScompT) :: dssComp
    integer         :: filtNum
    type(DSfilterT) :: filter
    character(dssFiltCharLength) :: tmpDssFiltName

    !-------------------------------------------------------------------
    ! Start:

    dssComp = dssComponent(dssPointer, compNum)

    ! from the DEBUG section in getNum1stDssCompWithValue()
        gtiBlockName = ''
        do filtNum = 0, dssFilters(dssComp) - 1
          filter = dssFilter(dssComp, filtNum)
          tmpDssFiltName = dssFilterName(filter)
          if (trim(tmpDssFiltName) .eq. 'TIME') then
            gtiBlockName = dssFilterBlock(filter)
            exit
          end if
        end do

  end subroutine getGtiBlockName


end program testdss_aux02
