program testregridders1
  use errorhandling
  use types
  use dal		! For INTEGER_NULL
  !use cal
  !use dss
  !use dataFilter	! in selectlib.

  !! Followings are in ssclib
  use array_utils	! for aryStatInfoSingleT and getAryStatInfo()
  !use cal_aux		! for getCalInfo()
  !use dal_aux		! for readColDataToFixed()
  use constants		! for Pi
  !use coordinate
  !use ellipse		! Called from Library source_cutouts
  use geometric_types	! for Point2dT type
  use regridders
  use test_utils	! for isNearlyEqual()
  !use wcs_aux

  implicit none

  call test_calcOldPixelCornersAffine
  call test_regridCartesian

contains

  !---------------------------------------------------------------------
  subroutine test_calcOldPixelCornersAffine
  !---------------------------------------------------------------------
  ! Tests of calcOldPixelCornersAffine() in regriders
  !

    implicit none

    character(30)  :: errHead
    real(single),   allocatable :: oldImage(:,:), newImage(:,:)
    type(Point2dT), allocatable :: oldPixelCorners(:,:)
    real(single) :: lEdge(2), uEdge(2)
    type(AryEdgesInfoT) :: oldAryEdgesInfo, newAryEdgesInfo	! defined in array_utils
    real(double) :: mtrxLinTrans(2,2), vecTranslate(2)
    integer :: mtrxNormal(2,2)
    ! integer, parameter :: mtrxNormal(2,2) = (/ (/1, 0/), (/0, 1/) /)	! invalid format...?

    ! integer :: status
    ! integer :: constVal
    ! integer :: i, j, lboundOldImage, uboundOldImage
    ! type(aryStatInfoSingleT) :: oldStatInfo, newStatInfo	! defined in array_utils

    !-------------------------------------------------------------------
    ! Start:

    mtrxNormal(1,:) = (/ 1, 0 /)
    mtrxNormal(2,:) = (/ 0, 1 /)

    !----------------------------------------------------------
    errHead = "CalcOldPixelCornersAffine001a"
    print *,"=== "//trim(errHead)//": Identical."

    allocate(oldImage(5,5), newImage(5,5), oldPixelCorners(6,6))
    oldAryEdgesInfo = getAryEdgesInfo(oldImage, style='wcs')
    newAryEdgesInfo = getAryEdgesInfo(newImage, style='wcs')
    mtrxLinTrans = mtrxNormal
    vecTranslate = 0.0

    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldAryEdgesInfo, newAryEdgesInfo, mtrxLinTrans, vecTranslate)

    if ( oldPixelCorners(1,1)%x .ne. 0.5  .or. &
         oldPixelCorners(1,1)%y .ne. 0.5  .or. &
         oldPixelCorners(3,1)%x .ne. 2.5  .or. &
         oldPixelCorners(3,1)%y .ne. 0.5  .or. &
         oldPixelCorners(1,3)%x .ne. 0.5  .or. &
         oldPixelCorners(1,3)%y .ne. 2.5 ) then
      print *, 'oldPixelCorners(:,1)%x =',oldPixelCorners(:,1)%x
      print *, 'oldPixelCorners(:,1)%y =',oldPixelCorners(:,1)%y
      print *, 'oldPixelCorners(3,1)%x =',oldPixelCorners(3,1)%x
      call error("err"//trim(errHead), '')
    end if

    deallocate(oldImage, newImage, oldPixelCorners)

    !----------------------------------------------------------
    errHead = "CalcOldPixelCornersAffine002a"
    print *,"=== "//trim(errHead)//": Identical but old/newImage starts from non-1."

    allocate(oldImage(0:4,-1:3), newImage(0:4,-1:3), oldPixelCorners(6,6))
    oldAryEdgesInfo = getAryEdgesInfo(oldImage, lboundIndex=lbound(oldImage), style='wcs')
    newAryEdgesInfo = getAryEdgesInfo(newImage, lboundIndex=lbound(newImage), style='wcs')
    mtrxLinTrans = mtrxNormal
    vecTranslate = 0.0

    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldAryEdgesInfo, newAryEdgesInfo, mtrxLinTrans, vecTranslate)

    if ( oldPixelCorners(1,1)%x .ne. 0.5  .or. &
         oldPixelCorners(1,1)%y .ne. 0.5  .or. &
         oldPixelCorners(3,1)%x .ne. 2.5  .or. &
         oldPixelCorners(3,1)%y .ne. 0.5  .or. &
         oldPixelCorners(1,3)%x .ne. 0.5  .or. &
         oldPixelCorners(1,3)%y .ne. 2.5 ) then	! Identical as Test 001a.
      call printAryEdgesInfo(oldAryEdgesInfo, 'oldAryEdgesInfo')
      print *, 'oldPixelCorners(:,1)%x =',oldPixelCorners(:,1)%x
      print *, 'oldPixelCorners(:,1)%y =',oldPixelCorners(:,1)%y
      print *, 'oldPixelCorners(3,1)%x =',oldPixelCorners(3,1)%x
      call error("err"//trim(errHead), '')
    end if

    deallocate(oldImage, newImage, oldPixelCorners)

    !----------------------------------------------------------
    errHead = "CalcOldPixelCornersAffine003a"
    print *,"=== "//trim(errHead)//": No physical transform, but with a newImage with a bigger coverage."

    allocate(oldImage(5,5), newImage(-1:6,0:8), oldPixelCorners(6,6))
    oldAryEdgesInfo = getAryEdgesInfo(oldImage, style='wcs')
    newAryEdgesInfo = getAryEdgesInfo(newImage, lboundIndex=lbound(newImage), style='wcs')
    ! Therefore, the indices of 3rd and 2nd for X and Y, respectively,
    ! in newImage correspond to (1,1) in oldImage.
    mtrxLinTrans = mtrxNormal
    vecTranslate = 0.0

    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldAryEdgesInfo, newAryEdgesInfo, mtrxLinTrans, vecTranslate)

    if ( oldPixelCorners(1,1)%x .ne. 2.5  .or. &
         oldPixelCorners(1,1)%y .ne. 1.5  .or. &
         oldPixelCorners(3,1)%x .ne. 4.5  .or. &
         oldPixelCorners(3,1)%y .ne. 1.5  .or. &
         oldPixelCorners(1,3)%x .ne. 2.5  .or. &
         oldPixelCorners(1,3)%y .ne. 3.5 ) then
      call printAryEdgesInfo(newAryEdgesInfo, 'newAryEdgesInfo')
      print *, 'oldPixelCorners(:,1)%x =',oldPixelCorners(:,1)%x
      print *, 'oldPixelCorners(:,1)%y =',oldPixelCorners(:,1)%y
      print *, 'oldPixelCorners(3,1)%x =',oldPixelCorners(3,1)%x
      call error("err"//trim(errHead), '')
    end if

    deallocate(oldImage, newImage, oldPixelCorners)

    !----------------------------------------------------------
    errHead = "CalcOldPixelCornersAffine004a"
    print *,"=== "//trim(errHead)//": No physical transform, but with a newImage with a much bigger pixel size."

    allocate(oldImage(5,5), newImage(1:1,1:1), oldPixelCorners(6,6))
    oldAryEdgesInfo = getAryEdgesInfo(oldImage, style='wcs')
    lEdge = 0.5
    uEdge = 5.5
    newAryEdgesInfo = getAryEdgesInfo(newImage, lEdge, uEdge, lboundIndex=lbound(newImage))
    mtrxLinTrans = mtrxNormal
    vecTranslate = 0.0

    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldAryEdgesInfo, newAryEdgesInfo, mtrxLinTrans, vecTranslate)

    if ( oldPixelCorners(1,1)%x .ne. 0.5  .or. &
         oldPixelCorners(1,1)%y .ne. 0.5  .or. &
         oldPixelCorners(2,1)%x .ne. 0.7  .or. &
         oldPixelCorners(6,1)%x .ne. 1.5  .or. &
         oldPixelCorners(6,1)%y .ne. 0.5  .or. &
         oldPixelCorners(1,6)%x .ne. 0.5  .or. &
         oldPixelCorners(1,6)%y .ne. 1.5 ) then
      call printAryEdgesInfo(newAryEdgesInfo, 'newAryEdgesInfo')
      print *, 'oldPixelCorners(:,1)%x =',oldPixelCorners(:,1)%x
      print *, 'oldPixelCorners(:,1)%y =',oldPixelCorners(:,1)%y
      print *, 'oldPixelCorners(1,6)%y =',oldPixelCorners(1,6)%y
      call error("err"//trim(errHead), '')
    end if

    deallocate(oldImage, newImage, oldPixelCorners)

    !----------------------------------------------------------
    errHead = "CalcOldPixelCornersAffine005a"
    print *,"=== "//trim(errHead)//": No physical transform, "&
       //"but with a newImage with a much bigger pixel size and wider area coverage."

    allocate(oldImage(5,5), newImage(1:3,1:3), oldPixelCorners(6,6))
    oldAryEdgesInfo = getAryEdgesInfo(oldImage, style='wcs')
    lEdge = 0.5-5
    uEdge = 5.5+5
    newAryEdgesInfo = getAryEdgesInfo(newImage, lEdge, uEdge, lboundIndex=lbound(newImage))
    mtrxLinTrans = mtrxNormal
    vecTranslate = 0.0

    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldAryEdgesInfo, newAryEdgesInfo, mtrxLinTrans, vecTranslate)

    if ( oldPixelCorners(1,1)%x .ne. 1.5  .or. &
         oldPixelCorners(1,1)%y .ne. 1.5  .or. &
         oldPixelCorners(2,1)%x .ne. 1.7  .or. &
         oldPixelCorners(6,1)%x .ne. 2.5  .or. &
         oldPixelCorners(6,1)%y .ne. 1.5  .or. &
         oldPixelCorners(1,6)%x .ne. 1.5  .or. &
         oldPixelCorners(1,6)%y .ne. 2.5 ) then
      call printAryEdgesInfo(newAryEdgesInfo, 'newAryEdgesInfo')
      print *, 'oldPixelCorners(:,1)%x =',oldPixelCorners(:,1)%x
      print *, 'oldPixelCorners(:,1)%y =',oldPixelCorners(:,1)%y
      print *, 'oldPixelCorners(1,6)%y =',oldPixelCorners(1,6)%y
      call error("err"//trim(errHead), '')
    end if

    deallocate(oldImage, newImage, oldPixelCorners)

    !----------------------------------------------------------
    errHead = "CalcOldPixelCornersAffine011a"
    print *,"=== "//trim(errHead)//": Parallel move by x=+1, but the edges change so no change computationally."

    allocate(oldImage(0:4,-1:3), newImage(1:5,-1:3), oldPixelCorners(6,6))
    oldAryEdgesInfo = getAryEdgesInfo(oldImage, lboundIndex=lbound(oldImage), style='wcs')
    newAryEdgesInfo = getAryEdgesInfo(newImage, lboundIndex=lbound(newImage), style='wcs')
    mtrxLinTrans = mtrxNormal
    vecTranslate = (/ 1.0, 0.0 /)

    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldAryEdgesInfo, newAryEdgesInfo, mtrxLinTrans, vecTranslate)

    if ( oldPixelCorners(1,1)%x .ne. 0.5  .or. &
         oldPixelCorners(1,1)%y .ne. 0.5  .or. &
         oldPixelCorners(3,1)%x .ne. 2.5  .or. &
         oldPixelCorners(3,1)%y .ne. 0.5  .or. &
         oldPixelCorners(1,3)%x .ne. 0.5  .or. &
         oldPixelCorners(1,3)%y .ne. 2.5 ) then	! Identical as Test 001a.
      call printAryEdgesInfo(oldAryEdgesInfo, 'oldAryEdgesInfo')
      print *, 'oldPixelCorners(:,1)%x =',oldPixelCorners(:,1)%x
      print *, 'oldPixelCorners(:,1)%y =',oldPixelCorners(:,1)%y
      print *, 'oldPixelCorners(3,1)%x =',oldPixelCorners(3,1)%x
      call error("err"//trim(errHead), '')
    end if

    deallocate(oldImage, newImage, oldPixelCorners)

    !----------------------------------------------------------
    errHead = "CalcOldPixelCornersAffine012a"
    print *,"=== "//trim(errHead)//": Parallel move by x=+1."

    allocate(oldImage(5,5), newImage(0:6,0:6), oldPixelCorners(6,6))
    oldAryEdgesInfo = getAryEdgesInfo(oldImage, style='wcs')
    newAryEdgesInfo = getAryEdgesInfo(newImage, lboundIndex=lbound(newImage), style='wcs')
    mtrxLinTrans = mtrxNormal
    vecTranslate = (/ 1.0, 0.0 /)

    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldAryEdgesInfo, newAryEdgesInfo, mtrxLinTrans, vecTranslate)

    if ( oldPixelCorners(1,1)%x .ne. 2.5  .or. &
         oldPixelCorners(1,1)%y .ne. 1.5  .or. &
         oldPixelCorners(3,1)%x .ne. 4.5  .or. &
         oldPixelCorners(3,1)%y .ne. 1.5  .or. &
         oldPixelCorners(1,3)%x .ne. 2.5  .or. &
         oldPixelCorners(1,3)%y .ne. 3.5 ) then
      call printAryEdgesInfo(newAryEdgesInfo, 'newAryEdgesInfo')
      print *, 'oldPixelCorners(:,1)%x =',oldPixelCorners(:,1)%x
      print *, 'oldPixelCorners(:,1)%y =',oldPixelCorners(:,1)%y
      print *, 'oldPixelCorners(1,3)%y =',oldPixelCorners(1,3)%y
      call error("err"//trim(errHead), '')
    end if

    deallocate(oldImage, newImage, oldPixelCorners)

    !----------------------------------------------------------
    errHead = "CalcOldPixelCornersAffine013a"
    print *,"=== "//trim(errHead)//": Parallel move by y=+1."

    allocate(oldImage(5,5), newImage(0:6,0:6), oldPixelCorners(6,6))
    oldAryEdgesInfo = getAryEdgesInfo(oldImage, style='wcs')
    newAryEdgesInfo = getAryEdgesInfo(newImage, lboundIndex=lbound(newImage), style='wcs')
    mtrxLinTrans = mtrxNormal
    vecTranslate = (/ 0.0, 1.0 /)

    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldAryEdgesInfo, newAryEdgesInfo, mtrxLinTrans, vecTranslate)

    if ( oldPixelCorners(1,1)%x .ne. 1.5  .or. &
         oldPixelCorners(1,1)%y .ne. 2.5  .or. &
         oldPixelCorners(3,1)%x .ne. 3.5  .or. &
         oldPixelCorners(3,1)%y .ne. 2.5  .or. &
         oldPixelCorners(1,3)%x .ne. 1.5  .or. &
         oldPixelCorners(1,3)%y .ne. 4.5 ) then
      call printAryEdgesInfo(newAryEdgesInfo, 'newAryEdgesInfo')
      print *, 'oldPixelCorners(:,1)%x =',oldPixelCorners(:,1)%x
      print *, 'oldPixelCorners(:,1)%y =',oldPixelCorners(:,1)%y
      print *, 'oldPixelCorners(1,3)%y =',oldPixelCorners(1,3)%y
      call error("err"//trim(errHead), '')
    end if

    deallocate(oldImage, newImage, oldPixelCorners)

    !----------------------------------------------------------
    errHead = "CalcOldPixelCornersAffine021a"
    print *,"=== "//trim(errHead)//": Scaling by x=*2-0.5, but the edges change so no change computationally."

    allocate(oldImage(-1:3,-1:3), newImage(1:5,1:5), oldPixelCorners(6,6))
    lEdge = 0.5
    uEdge = 5.5
    oldAryEdgesInfo = getAryEdgesInfo(oldImage, lEdge, uEdge, lboundIndex=lbound(oldImage))
    ! getAryEdgesInfo(oldImage, style='wcs') for oldImage(1:5,1:5) would be the same.
    lEdge = 0.5
    uEdge = (/ 10.5, 5.5 /)
    newAryEdgesInfo = getAryEdgesInfo(newImage, lEdge, uEdge, lboundIndex=lbound(newImage))
    mtrxLinTrans = mtrxNormal
    mtrxLinTrans(1,1) = 2.0
    vecTranslate = (/ -0.5, 0.0 /)

    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldAryEdgesInfo, newAryEdgesInfo, mtrxLinTrans, vecTranslate)

    if ( oldPixelCorners(1,1)%x .ne. 0.5  .or. &
         oldPixelCorners(1,1)%y .ne. 0.5  .or. &
         oldPixelCorners(3,1)%x .ne. 2.5  .or. &
         oldPixelCorners(3,1)%y .ne. 0.5  .or. &
         oldPixelCorners(1,3)%x .ne. 0.5  .or. &
         oldPixelCorners(1,3)%y .ne. 2.5 ) then
      call printAryEdgesInfo(oldAryEdgesInfo, 'oldAryEdgesInfo')
      print *, 'oldPixelCorners(:,1)%x =',oldPixelCorners(:,1)%x
      print *, 'oldPixelCorners(:,1)%y =',oldPixelCorners(:,1)%y
      print *, 'oldPixelCorners(3,1)%x =',oldPixelCorners(3,1)%x
      call error("err"//trim(errHead), '')
    end if

    deallocate(oldImage, newImage, oldPixelCorners)

    !----------------------------------------------------------
    errHead = "CalcOldPixelCornersAffine022a"
    print *,"=== "//trim(errHead)//": Scaling by x=*2, but the edges change so no change computationally."

    allocate(oldImage(1:5,1:5), newImage(1:11,1:11), oldPixelCorners(6,6))
    oldAryEdgesInfo = getAryEdgesInfo(oldImage, style='wcs')
    newAryEdgesInfo = getAryEdgesInfo(newImage, style='wcs')
    mtrxLinTrans = mtrxNormal
    mtrxLinTrans(1,1) = 2.0
    vecTranslate = 0

    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldAryEdgesInfo, newAryEdgesInfo, mtrxLinTrans, vecTranslate)

    if ( oldPixelCorners(1,1)%x .ne. 1.0  .or. &
         oldPixelCorners(1,1)%y .ne. 0.5  .or. &
         oldPixelCorners(3,1)%x .ne. 5.0  .or. &
         oldPixelCorners(3,1)%y .ne. 0.5  .or. &
         oldPixelCorners(1,3)%x .ne. 1.0  .or. &
         oldPixelCorners(1,3)%y .ne. 2.5 ) then
      call printAryEdgesInfo(newAryEdgesInfo, 'newAryEdgesInfo')
      print *, 'oldPixelCorners(:,1)%x =',oldPixelCorners(:,1)%x
      print *, 'oldPixelCorners(:,1)%y =',oldPixelCorners(:,1)%y
      print *, 'oldPixelCorners(3,1)%x =',oldPixelCorners(3,1)%x
      call error("err"//trim(errHead), '')
    end if

    deallocate(oldImage, newImage, oldPixelCorners)

    !----------------------------------------------------------
    errHead = "CalcOldPixelCornersAffine031a"
    print *,"=== "//trim(errHead)//": Rotation by t=90degrees."

    allocate(oldImage(1:5,1:5), newImage(-5:5,-5:5), oldPixelCorners(6,6))
    oldAryEdgesInfo = getAryEdgesInfo(oldImage, style='wcs')
    newAryEdgesInfo = getAryEdgesInfo(newImage, lboundIndex=lbound(newImage), style='wcs')
    mtrxLinTrans(1,:) = (/ 0, -1 /)	! (cos t, -sin t), where t=90degrees
    mtrxLinTrans(2,:) = (/ 1,  0 /)	! (sin t,  cos t)  ie., normal rotation
    vecTranslate = 0

    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldAryEdgesInfo, newAryEdgesInfo, mtrxLinTrans, vecTranslate)

    if ( oldPixelCorners(1,1)%x .ne.  5.5  .or. &	! x'=-0.5
         oldPixelCorners(1,1)%y .ne.  6.5  .or. &	! y'=+0.5
         oldPixelCorners(6,1)%x .ne.  5.5  .or. &	! x'=-0.5
         oldPixelCorners(6,1)%y .ne. 11.5  .or. &	! y'=+5.5
         oldPixelCorners(1,6)%x .ne.  0.5  .or. &	! x'=-5.5
         oldPixelCorners(1,6)%y .ne.  6.5 ) then	! y'=+0.5
      call printAryEdgesInfo(newAryEdgesInfo, 'newAryEdgesInfo')
      print *, 'oldPixelCorners(:,1)%x =',oldPixelCorners(:,1)%x
      print *, 'oldPixelCorners(:,1)%y =',oldPixelCorners(:,1)%y
      print *, 'oldPixelCorners(1,6)%x =',oldPixelCorners(1,6)%x
      call error("err"//trim(errHead), '')
    end if

    deallocate(oldImage, newImage, oldPixelCorners)

    !----------------------------------------------------------
    errHead = "CalcOldPixelCornersAffine032a"
    print *,"=== "//trim(errHead)//": Rotation by t=90degrees but onto the translated frame."

    allocate(oldImage(1:5,1:5), newImage(-5:-1,1:5), oldPixelCorners(6,6))
    oldAryEdgesInfo = getAryEdgesInfo(oldImage, style='wcs')
    newAryEdgesInfo = getAryEdgesInfo(newImage, lboundIndex=lbound(newImage), style='wcs')
    mtrxLinTrans(1,:) = (/ 0, -1 /)	! (cos t, -sin t), where t=90degrees
    mtrxLinTrans(2,:) = (/ 1,  0 /)	! (sin t,  cos t)  ie., normal rotation
    vecTranslate = 0

    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldAryEdgesInfo, newAryEdgesInfo, mtrxLinTrans, vecTranslate)

    if ( oldPixelCorners(1,1)%x .ne. 5.5  .or. &
         oldPixelCorners(1,1)%y .ne. 0.5  .or. &
         oldPixelCorners(6,1)%x .ne. 5.5  .or. &
         oldPixelCorners(6,1)%y .ne. 5.5  .or. &
         oldPixelCorners(1,6)%x .ne. 0.5  .or. &
         oldPixelCorners(1,6)%y .ne. 0.5 ) then
      call printAryEdgesInfo(newAryEdgesInfo, 'newAryEdgesInfo')
      print *, 'oldPixelCorners(:,1)%x =',oldPixelCorners(:,1)%x
      print *, 'oldPixelCorners(:,1)%y =',oldPixelCorners(:,1)%y
      print *, 'oldPixelCorners(1,6)%x =',oldPixelCorners(1,6)%x
      call error("err"//trim(errHead), '')
    end if

    deallocate(oldImage, newImage, oldPixelCorners)

    !----------------------------------------------------------
    errHead = "CalcOldPixelCornersAffine041a"
    print *,"=== "//trim(errHead)//": Reflection to X-axis onto the translated frame."

    allocate(oldImage(1:5,1:5), newImage(-5:-1,1:5), oldPixelCorners(6,6))
    oldAryEdgesInfo = getAryEdgesInfo(oldImage, style='wcs')
    newAryEdgesInfo = getAryEdgesInfo(newImage, lboundIndex=lbound(newImage), style='wcs')
    mtrxLinTrans = mtrxNormal
    mtrxLinTrans(1,1) = -1.0

    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldAryEdgesInfo, newAryEdgesInfo, mtrxLinTrans, vecTranslate)

    if ( oldPixelCorners(1,1)%x .ne. 5.5  .or. &
         oldPixelCorners(1,1)%y .ne. 0.5  .or. &
         oldPixelCorners(6,1)%x .ne. 0.5  .or. &
         oldPixelCorners(6,1)%y .ne. 0.5  .or. &
         oldPixelCorners(1,6)%x .ne. 5.5  .or. &
         oldPixelCorners(1,6)%y .ne. 5.5 ) then
      call printAryEdgesInfo(newAryEdgesInfo, 'newAryEdgesInfo')
      print *, 'oldPixelCorners(:,1)%x =',oldPixelCorners(:,1)%x
      print *, 'oldPixelCorners(:,1)%y =',oldPixelCorners(:,1)%y
      print *, 'oldPixelCorners(1,6)%x =',oldPixelCorners(1,6)%x
      call error("err"//trim(errHead), '')
    end if

    deallocate(oldImage, newImage, oldPixelCorners)

  end subroutine test_calcOldPixelCornersAffine

  !---------------------------------------------------------------------
  subroutine test_regridCartesian
  !---------------------------------------------------------------------
  ! Tests of regridCartesian() in regriders
  !
  !
   ! ********* Arguments (oldPixelCorners)
   ! 
   ! In the conversion of arrays oldImage into newImage,
   ! suppose the oldImage arrays are given like this
   ! (the lbound-s of the array oldImage for (i,j)=(1,1),
   ! where i and j are indices of the array):
   ! 
   !      |   |   | 
   !  2.5 +---+---+-
   !      |   |   |
   ! j=2  |   |   |
   !      |   |   |
   !  1.5 +---+---+-
   !      |   |   |
   ! j=1  |   |   |
   !      |   |   |
   !  0.5 +---+---+-
   !     0.5 1.5 2.5
   !       i=1 i=2 i=3
   !  (k=1 k=2 k=3 k=4; an example of newImage)
   ! 
   ! It is assumed to be the "wcs pixel system".  Namely,
   ! the centre of the array oldImage (i,j)=(1,1) is defined
   ! as (x,y)=(1.0,1.0) where both x and y are expressed in the "i and j"
   ! index coordinates, therefore this pixel has the edges of
   !  Lower-Left:  (x,y)=(0.5, 0.5) 
   !  Lower-Right: (x,y)=(1.5, 0.5) 
   !  Upper-Left:  (x,y)=(0.5, 1.5) 
   !  Upper-Right: (x,y)=(1.5, 1.5) 
   ! The edges of the latter two are obviously common as
   ! for the pixel (i,j)=(1,2), and as such.
   ! 
   ! In oldPixelCorners, this relation is expressed, that is,
   !  oldPixelCorners(a,b)%x,y is the corresponding position of (x',y') on newImage
   !  which corresponds to the position of the (x,y) in oldImage.
   ! For example, if there is no conversion,
   !  oldPixelCorners(1,1)%x,y = (0.5, 0.5)
   !  oldPixelCorners(2,1)%x,y = (1.5, 0.5)
   !  oldPixelCorners(1,2)%x,y = (0.5, 1.5) etc.
   ! For example, if the newImage coordiates shift to the negative along X by 1,
   ! that is, the (x,y) in oldImage is now (x+1,y) in newImage
   ! (e.g., (1,1) => (2,1)),
   !  oldPixelCorners(1,1)%x,y = (1.5, 0.5)
   !  oldPixelCorners(2,1)%x,y = (2.5, 0.5)
   !  oldPixelCorners(1,2)%x,y = (1.5, 1.5) etc.
   ! 
   ! As the edges of each pixel are shared with its adjascents,
   ! the dimensions of oldPixelCorners are larger by just one
   ! than oldImage (nb., the dimensions of newImage are irrelevant).
   ! 
   ! In reality, (i,j) can physically mean anything, such as,
   ! (i,j)=(1,1) is (RA,DEC)=(12.3, -45.6) etc.
   ! However, this routine does not consider such physical meaning of the axes
   ! at all.
   ! 
   ! ********* Background
   ! 
   ! (** Case 1: Parallel move to X direction **)
   ! 
   ! Transform of C1(x) -> D1(x-2)
   !   i.e., x_D1 = x_C1 - 2   (e.g., x_C1=2 => x_D1=0)
   !         y_D1 = y_C1
   ! When the initial C1 is given as,
   !   where (x=1) then
   !     C1=7 
   !   elsewhere
   !     C1=0 
   !   end where
   ! the converted D1 must be as
   !   where (x=-1) then
   !     D1=7 
   !   elsewhere
   !     D1=0 
   !   end where
   ! 
   ! This is expressed in the following drawing,
   ! 
   !           | (Values)            |   |      
   !   0 0    4| 7 0 0 0         0  4| 7   0 0 
   !           |                     |   |      
   !   0 0    3| 7 0 0 0         0  3| 7   0 0 
   !           |                     |   |      
   !   0 0    2| 7 0 0 0         0  2| 7   0 0 
   !           |                     |   |      
   !   0 0    1| 7 0 0 0         0  1| 7   0 0 
   !           |                     |   |      
   !  -+-0---+-+-7-0-+-+-   -+-+-+-+-+-7-+-+------- 
   !  -4-3  -1 0 1 2 3 4    -4-3  -1 0 1 2 3 4      C1
   !           |                -4-3  -1 0 1 2 3 4  D1
   !   0 0   -1+ 7 0 0 0         0 -1+ 7   0 0
   !           |                     |   |      
   !   0 0   -2+ 7 0 0 0         0 -2+ 7   0 0 
   !           |                     |   |      
   !   0 0   -3+ 7 0 0 0         0 -3+ 7   0 0 
   !           |                     |   |      
   !   0 0   -4+ 7 0 0 0         0 -4+ 7   0 0 
   !           |                     |   |      
   !         (C1)                  (C1) (D1)             
   ! 
   ! 
   ! (** Case 2: Parallel move to X direction **)
   ! 
   ! Transform of C1(y) -> D1(y-2)
   !   i.e., x_D1 = x_C1
   !         y_D1 = y_C1 - 2   (e.g., y_C1=2 => y_D1=0)
   ! When the initial C1 is given as,
   !   where (y=1) then
   !     C1=7 
   !   elsewhere
   !     C1=0 
   !   end where
   ! the converted D1 must be as
   !   where (y=-1) then
   !     D1=7 
   !   elsewhere
   !     D1=0 
   !   end where
   ! 
   ! This is expressed in the following drawing,
   ! 
   !           | (Values)            |          
   !   0 0    4| 0 0 0 0         0  4|2  0 0 0 
   !           |                     |          
   !   0 0    3| 0 0 0 0     0 0    3|1  0 0 0     
   !           |                     |          
   !   0 0    2| 0 0 0 0     - - - -2+0- - - - -  
   !           |                     |          
   !   7 7 7 717 7 7 7 7     7 7 7  1-1  7 7 7 
   !           |                     |          
   !  -+-0---+-+---0-+-+-   -+-+-+-+-+---+-+------- 
   !  -4-3  -1 0 1 2 3 4    -4-3  -1 0 1 2 3 4     (C1:D1)
   !           |                     |
   !   0 0   -1+ 0 0 0 0     0 0 0 -1| 0   0 0
   !           |                     |          
   !   0 0   -2+ 0 0 0 0     0 0 0 -2-4  0 0 0 
   !           |                     |          
   !   0 0   -3+ 0 0 0 0     0 0 0 -3+ 0 0 0 0 
   !           |                     |          
   !   0 0   -4+ 0 0 0 0     0 0 0 -4-6  0 0 0 
   !           |                     |         
   !         (C1)                 (C1:D1)            
   !
   !
   ! (** Case 3: Coorinate system itself is rotated by 45 degree clockwise **)
   ! 
   ! Transform of C1(x,y) -> D1(x,y)
   !  (x_D1)    ( cos_t  sin_t) (x_C1)
   !  (    ) =  (             ) (    )
   !  (y_D1)    (-sin_t  cos_t) (y_C1)
   ! 
   !   eg. (1, 0) => (  sqrt(2)/2,  sqrt(2)/2)
   !       (0, 1) => ( -sqrt(2)/2,  sqrt(2)/2)
   !       (3.5, 3.5) => (5, 0)  where 5*sqrt(2)/2 = 5/sqrt(2) =~ 3.5
   !
   ! When the initial C1 is given as,
   !   where (x=y) then
   !     C1=7 
   !   elsewhere
   !     C1=0 
   !   end where
   ! the converted D1 must be as
   !   where (y=0) then
   !     D1=7 
   !   elsewhere
   !     D1=0 
   !   end where
   ! Note that the covered areas between them for the limited area are different.
   ! 
   ! This is expressed in the following drawing,
   ! 
   !           | (Values)  
   !   0 0   y4| 0 0 0 7(D1:x) 
   !  y5\      |      /5  
   !     \   y3| 0 0 7    
   !      \    |    /     
   !   0 0 \ y2| 0 7 0 0  
   !      y2\  |  /2      
   !   0 0 0 \1| 7   0 0  
   !          \|/         
   !  -+-0---+-*---0-+-+- 
   !  -4-3  -1 0 1 2 3 4  (C1:x)
   !           |\         
   !   0 0   -1+ \   0 0  
   !      -2/  |-2\y      
   !   0 0 7 -2+y  \ 0 0  
   !      /    |    \     
   !   0 7   -3+y  0 \ 0  
   !  -5/      |    -5\y  
   !   7 0   -4+y  0 0 \  
   !           |       (D1:y)
   !         (C1:y)         
   ! 
   ! 
   ! (** Case 4: Reflection (mirror image) **)
   ! 
   ! 
   ! 

    implicit none

    character(25)  :: errHead
    real(single),   allocatable :: oldImage(:,:)
    type(Point2dT), allocatable :: oldPixelCorners(:,:)
    real(single),   allocatable :: newImage(:,:)
    integer :: status
    logical(bool) :: testFlagArg
    character(20) :: averagingStyle	! The number "20" is based on localAveragingStyle in regridVector()
    integer :: constVal
    integer :: i, j, lboundOldImage, uboundOldImage, iSize
    type(aryStatInfoSingleT) :: oldStatInfo, newStatInfo	! defined in array_utils
    type(AryEdgesInfoT) :: oldImageEdgesInfo, newImageEdgesInfo	! defined in array_utils
    real(double) :: a
    real(double),   dimension(2,2) :: mtrxLinTrans
    real(double),   dimension(2)   :: vecTranslate
    real(single),   dimension(2)   :: lEdge, uEdge
    integer,        dimension(2,2) :: mtrxNormal
    integer(int32), dimension(2)   :: oldIndices, newIndices
    integer, parameter :: maxPrecision = 5

    !-------------------------------------------------------------------
    ! Start:

    averagingStyle = 'rate'
    testFlagArg = .false.
    mtrxNormal(1,:) = (/ 1, 0 /)	! parameter (=constant)
    mtrxNormal(2,:) = (/ 0, 1 /)	! parameter (=constant)

    lboundOldImage = 1
    uboundOldImage = 9
    allocate(oldImage(lboundOldImage:uboundOldImage, lboundOldImage:uboundOldImage))
    allocate(newImage(lboundOldImage:uboundOldImage, lboundOldImage:uboundOldImage))
    oldImageEdgesInfo = getAryEdgesInfo(oldImage, lboundIndex=lbound(oldImage), style='wcs')
    newImageEdgesInfo = getAryEdgesInfo(newImage, lboundIndex=lbound(newImage), style='wcs')

    !----------------------------------------------------------
    errHead = "RegridCartesian001a"
    print *,"=== "//trim(errHead)//": No conversion. Delta function. Crude way."

    constVal = 1000.0
    oldImage(5,5) = constVal	! Delta function.

    ! if (.false.) then	! if oldPixelCorners(0:9)
    ! allocate(oldPixelCorners( &
    !    lboundOldImage-1:uboundOldImage, &
    !    lboundOldImage-1:uboundOldImage) )	! nb.: the size is larger by 1.
    ! do i=lboundOldImage-1, uboundOldImage	! nb. starts from lbound-1
    !   do j=lboundOldImage-1, uboundOldImage
    !     oldPixelCorners(i,j)%x = i
    !     oldPixelCorners(i,j)%y = j
    !   end do
    ! end do
    ! else if (.false.) then		! if oldPixelCorners(1:10)
    ! allocate(oldPixelCorners( &
    !    lboundOldImage:uboundOldImage+1, &
    !    lboundOldImage:uboundOldImage+1) )	! nb.: the size is larger by 1.
    ! do i=lboundOldImage, uboundOldImage+1	! nb. starts from lbound-1
    !   do j=lboundOldImage, uboundOldImage+1
    !     oldPixelCorners(i,j)%x = i
    !     oldPixelCorners(i,j)%y = j
    !   end do
    ! end do
    ! else		! if oldPixelCorners(1:10) -- treating as the corners.
    allocate(oldPixelCorners( &
       lboundOldImage:uboundOldImage+1, &
       lboundOldImage:uboundOldImage+1) )	! nb.: the size is larger by 1.
    do i=lboundOldImage, uboundOldImage+1	! nb. starts from lbound-1
      do j=lboundOldImage, uboundOldImage+1
        oldPixelCorners(i,j)%x = i-0.5
        oldPixelCorners(i,j)%y = j-0.5
      end do
    end do
    ! end if

    newImage(:,:) = 0.0

    call regridCartesian(oldPixelCorners, oldImage, newImage, status&
       , testFlagArg, trim(averagingStyle))

    if ( status .ne. 0 ) then
      print *, 'status=',status
      call error("err"//trim(errHead), '')
    end if

    oldStatInfo = getAryStatInfo(oldImage)	! defined in array_utils
    newStatInfo = getAryStatInfo(newImage)	! defined in array_utils

    if ( oldStatInfo%realsum .ne. newStatInfo%realsum ) then
      print *, 'oldStatInfo%realsum, newStatInfo%realsum = ' &
              , oldStatInfo%realsum, newStatInfo%realsum
      call error("err"//trim(errHead), '')
    end if

!!print *,'oldPixelCorners%x=',oldPixelCorners%x
!!print *,'oldPixelCorners%y=',oldPixelCorners%y
!print *,'oldPixelCorners(1,1)%x,y=',oldPixelCorners(1,1)%x,oldPixelCorners(1,1)%y
!print *,'oldPixelCorners(2,2)%x,y=',oldPixelCorners(2,2)%x,oldPixelCorners(2,2)%y
!!print *,'oldPixelCorners(-2,-2)%x,y=',oldPixelCorners(-2,-2)%x,oldPixelCorners(-2,-2)%y
!!print *,'oldImage(-3,:)=',oldImage(-3,:)
!!print *,'newImage(-3,:)=',newImage(-3,:)
!print *,'oldImage(1,:)=',oldImage(1,:)
!print *,'oldImage(2,:)=',oldImage(2,:)
!print *,'oldImage(3,:)=',oldImage(3,:)
!print *,'oldImage(4,:)=',oldImage(4,:)
!print *,'oldImage(5,:)=',oldImage(5,:)
!print *,'oldImage(6,:)=',oldImage(6,:)
!print *,'oldImage(7,:)=',oldImage(7,:)
!print *,'oldImage(8,:)=',oldImage(8,:)
!print *,'oldImage(9,:)=',oldImage(9,:)
!print *,'newImage(1,:)=',newImage(1,:)
!print *,'newImage(2,:)=',newImage(2,:)
!print *,'newImage(3,:)=',newImage(3,:)
!print *,'newImage(4,:)=',newImage(4,:)
!print *,'newImage(5,:)=',newImage(5,:)
!print *,'newImage(6,:)=',newImage(6,:)
!print *,'newImage(7,:)=',newImage(7,:)
!print *,'newImage(8,:)=',newImage(8,:)
!print *,'newImage(9,:)=',newImage(9,:)
!!print *,'oldImage(:,1)=',oldImage(:,1)
!!print *,'newImage(:,1)=',newImage(:,1)

    if ( oldStatInfo%maxval .ne. newStatInfo%maxval ) then
      print *, 'newStatInfo%maxval = ', newStatInfo%maxval
      call error("err"//trim(errHead), '')
    end if

    ! if ( any(newImage(-1,:) .ne. constVal) ) then
    !   print *, 'newImage(-1,:) = ', newImage(-1,:)
    !   call error("err"//trim(errHead), '')
    ! end if
    ! 
    ! if ( any(newImage(2,2:8) .ne. 7.0) ) then
    !   print *, 'newImage(2,2:8) = ', newImage(2,2:8)
    !   call error("err"//trim(errHead), '')
    ! end if


    !----------------------------------------------------------
    errHead = "RegridCartesian002a"
    print *,"=== "//trim(errHead)//": No conversion. Delta function."

    constVal = 1000.0
    oldImage(5,5) = constVal	! Delta function.

    ! allocate(oldPixelCorners( &
    !    lboundOldImage:uboundOldImage+1, &
    !    lboundOldImage:uboundOldImage+1) )	! nb.: the size is larger by 1.
    oldPixelCorners(:,:)%x = -99.0
    oldPixelCorners(:,:)%y = -99.0
    mtrxLinTrans = mtrxNormal
    vecTranslate = 0.0
    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldImageEdgesInfo, newImageEdgesInfo, mtrxLinTrans, vecTranslate)

    newImage(:,:) = 0.0

    call regridCartesian(oldPixelCorners, oldImage, newImage, status&
       , testFlagArg, trim(averagingStyle))

    if ( status .ne. 0 ) then
      print *, 'status=',status
      call error("err"//trim(errHead), '')
    end if

    oldStatInfo = getAryStatInfo(oldImage)	! defined in array_utils
    newStatInfo = getAryStatInfo(newImage)	! defined in array_utils

    if ( oldStatInfo%realsum .ne. newStatInfo%realsum ) then
      print *, 'oldStatInfo%realsum, newStatInfo%realsum = ' &
              , oldStatInfo%realsum, newStatInfo%realsum
      call error("err"//trim(errHead), '')
    end if

    if ( oldStatInfo%maxval .ne. newStatInfo%maxval ) then
      print *, 'newStatInfo%maxval = ', newStatInfo%maxval
      call error("err"//trim(errHead), '')
    end if

    if ( newImage(5,5) .ne. constVal ) then
      print *, 'newImage(5,5) = ', newImage(5,5)
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "RegridCartesian003a"
    print *,"=== "//trim(errHead)//": No conversion. Uniform."

    constVal = 7.0
    oldImage(:,:) = constVal	! Uniform.
    newImage(:,:) = 0.0

    call regridCartesian(oldPixelCorners, oldImage, newImage, status&
       , testFlagArg, trim(averagingStyle))

    if ( status .ne. 0 ) then
      print *, 'status=',status
      call error("err"//trim(errHead), '')
    end if

    oldStatInfo = getAryStatInfo(oldImage)	! defined in array_utils
    newStatInfo = getAryStatInfo(newImage)	! defined in array_utils

    if (.not. isNearlyEqual(oldStatInfo%realsum, newStatInfo%realsum, maxPrecision)) then
    ! if ( oldStatInfo%realsum .ne. newStatInfo%realsum ) then
    !!  => oldStatInfo%realsum, newStatInfo%realsum =    5.6700000000000000E+02   5.6699991941452026E+02
      print *, 'oldStatInfo%realsum, newStatInfo%realsum = ' &
              , oldStatInfo%realsum, newStatInfo%realsum
!print *,'oldPixelCorners%x=',oldPixelCorners%x
!print *,'oldPixelCorners%y=',oldPixelCorners%y
print *,'oldPixelCorners(1,1)%x,y=',oldPixelCorners(1,1)%x,oldPixelCorners(1,1)%y
print *,'oldPixelCorners(2,2)%x,y=',oldPixelCorners(2,2)%x,oldPixelCorners(2,2)%y
!print *,'oldPixelCorners(-2,-2)%x,y=',oldPixelCorners(-2,-2)%x,oldPixelCorners(-2,-2)%y
!print *,'oldImage(-3,:)=',oldImage(-3,:)
!print *,'newImage(-3,:)=',newImage(-3,:)
print *,'oldImage(1,:)=',oldImage(1,:)
print *,'oldImage(2,:)=',oldImage(2,:)
print *,'oldImage(9,:)=',oldImage(9,:)
print *,'newImage(1,:)=',newImage(1,:)
print *,'newImage(2,:)=',newImage(2,:)
print *,'newImage(9,:)=',newImage(9,:)
print *,'oldImage(:,1)=',oldImage(:,1)
print *,'newImage(:,1)=',newImage(:,1)
      call error("err"//trim(errHead), '')
    end if

    if (.not. isNearlyEqual(oldStatInfo%maxval, newStatInfo%maxval, maxPrecision)) then
    ! if ( oldStatInfo%maxval .ne. newStatInfo%maxval ) then
      print *, 'oldStatInfo%maxval, newStatInfo%maxval = ', oldStatInfo%maxval, newStatInfo%maxval
      call error("err"//trim(errHead), '')
    end if

    if ( any(newImage(:,:) .ne. constVal) ) then	! most stringent test
      print *, 'All the points should be (ideally) ', constVal
      print *, 'However it may not be quite achieved in 64-bit machines.'
      print *, 'newImage(:,:) = ', newImage(:,:)
      call warning("warning"//trim(errHead), '')
      ! call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "RegridCartesian011a"
    print *,"=== "//trim(errHead)//": Parallel move for X (x2=x1-2). (Case 1)"

    deallocate(oldImage, newImage, oldPixelCorners)
    lboundOldImage = -4
    uboundOldImage = 4
    iSize = uboundOldImage - lboundOldImage + 1
    allocate(oldImage(lboundOldImage:uboundOldImage, lboundOldImage:uboundOldImage))
    allocate(newImage(lboundOldImage-2:uboundOldImage+2, lboundOldImage:uboundOldImage))
    allocate(oldPixelCorners(iSize+1, iSize+1))	! nb.: the size is larger by 1.
    oldImageEdgesInfo = getAryEdgesInfo(oldImage, lboundIndex=lbound(oldImage), style='wcs')
    newImageEdgesInfo = getAryEdgesInfo(newImage, lboundIndex=lbound(newImage), style='wcs')

    constVal = 7.0
    oldImage(:,:) = 0.0
    oldImage(1,:) = constVal	! The value is 7 if x=1, else 0.

    oldPixelCorners(:,:)%x = -99.0
    oldPixelCorners(:,:)%y = -99.0
    ! Transform of C1(x) -> D1(x-2)
    !   i.e., x_D1 = x_C1 - 2   (e.g., x_C1=2 => x_D1=0)
    !         y_D1 = y_C1
    mtrxLinTrans = mtrxNormal
    vecTranslate = (/ -2, 0 /)
    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldImageEdgesInfo, newImageEdgesInfo, mtrxLinTrans, vecTranslate)

    newImage(:,:) = 0.0

    call regridCartesian(oldPixelCorners, oldImage, newImage, status&
       , testFlagArg, trim(averagingStyle))

    if ( status .ne. 0 ) then
      print *, 'status=',status
      call error("err"//trim(errHead), '')
    end if

    oldStatInfo = getAryStatInfo(oldImage)	! defined in array_utils
    newStatInfo = getAryStatInfo(newImage)	! defined in array_utils

    if (.not. isNearlyEqual(oldStatInfo%realsum, newStatInfo%realsum, maxPrecision)) then
    ! if ( oldStatInfo%realsum .ne. newStatInfo%realsum ) then
      print *, 'oldStatInfo%realsum, newStatInfo%realsum = ' &
              , oldStatInfo%realsum, newStatInfo%realsum
!!print *,'oldPixelCorners%x=',oldPixelCorners%x
!!print *,'oldPixelCorners%y=',oldPixelCorners%y
!print *,'oldPixelCorners(1,1)%x,y=',oldPixelCorners(1,1)%x,oldPixelCorners(1,1)%y
!print *,'oldPixelCorners(2,2)%x,y=',oldPixelCorners(2,2)%x,oldPixelCorners(2,2)%y
!print *,'oldPixelCorners(-2,-2)%x,y=',oldPixelCorners(-2,-2)%x,oldPixelCorners(-2,-2)%y
!print *,'oldImage=',oldImage 
!print *,'newImage=',newImage 
      call error("err"//trim(errHead), '')
    end if

    if ( any(newImage(-1,:) .ne. constVal) ) then
      print *, 'All the points should be (ideally) ', constVal
      print *, 'However it may not be quite achieved in 64-bit machines.'
      print *, 'newImage(-1,:) = ', newImage(-1,:)
      call warning("warning"//trim(errHead), '')
    end if

    if ( any(newImage(1,:) .ne. 0.0) ) then
      print *, 'newImage(1,:) = ',  newImage(1,:)
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "RegridCartesian012a"
    print *,"=== "//trim(errHead)//": Parallel move for Y (y2=y1-2). (Case 2)"

    deallocate(oldImage, newImage, oldPixelCorners)
    lboundOldImage = -4
    uboundOldImage = 4
    iSize = uboundOldImage - lboundOldImage + 1
    allocate(oldImage(lboundOldImage:uboundOldImage, lboundOldImage:uboundOldImage))
    allocate(newImage(lboundOldImage:uboundOldImage, lboundOldImage-2:uboundOldImage+2))
    allocate(oldPixelCorners(iSize+1, iSize+1))	! nb.: the size is larger by 1.
    oldImageEdgesInfo = getAryEdgesInfo(oldImage, lboundIndex=lbound(oldImage), style='wcs')
    newImageEdgesInfo = getAryEdgesInfo(newImage, lboundIndex=lbound(newImage), style='wcs')

    constVal = 7.0
    oldImage(:,:) = 0.0
    oldImage(:,1) = constVal	! The value is 7 if x=1, else 0.

    oldPixelCorners(:,:)%x = -99.0
    oldPixelCorners(:,:)%y = -99.0
    ! Transform of C1(y) -> D1(y-2)
    !   i.e., x_D1 = x_C1
    !         y_D1 = y_C1 - 2   (e.g., y_C1=2 => y_D1=0)
    mtrxLinTrans = mtrxNormal
    vecTranslate = (/ 0, -2 /)
    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldImageEdgesInfo, newImageEdgesInfo, mtrxLinTrans, vecTranslate)

    newImage(:,:) = 0.0

    call regridCartesian(oldPixelCorners, oldImage, newImage, status&
       , testFlagArg, trim(averagingStyle))

    if ( status .ne. 0 ) then
      print *, 'status=',status
      call error("err"//trim(errHead), '')
    end if

    oldStatInfo = getAryStatInfo(oldImage)	! defined in array_utils
    newStatInfo = getAryStatInfo(newImage)	! defined in array_utils

    if (.not. isNearlyEqual(oldStatInfo%realsum, newStatInfo%realsum, maxPrecision)) then
    ! if ( oldStatInfo%realsum .ne. newStatInfo%realsum ) then
      print *, 'oldStatInfo%realsum, newStatInfo%realsum = ' &
              , oldStatInfo%realsum, newStatInfo%realsum
      call error("err"//trim(errHead), '')
    end if

    if ( any(newImage(:,-1) .ne. constVal) ) then
      print *, 'All the points should be (ideally) ', constVal
      print *, 'However it may not be quite achieved in 64-bit machines.'
      print *, 'newImage(-1,:) = ', newImage(-1,:)
      call warning("warning"//trim(errHead), '')
    end if

    if ( any(newImage(:,1) .ne. 0.0) ) then
      print *, 'newImage(1,:) = ',  newImage(1,:)
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "RegridCartesian021a"
    print *,"=== "//trim(errHead)//": Scaling along +X by 4."

    deallocate(oldImage, newImage, oldPixelCorners)
    iSize = 5
    allocate(oldImage(iSize,iSize), newImage(30,30))
    allocate(oldPixelCorners(iSize+1, iSize+1))	! nb.: the size is larger by 1.
    oldImageEdgesInfo = getAryEdgesInfo(oldImage, lboundIndex=lbound(oldImage), style='wcs')
    newImageEdgesInfo = getAryEdgesInfo(newImage, lboundIndex=lbound(newImage), style='wcs')

    constVal = 100.0
    oldImage(:,:) = 0.0
    oldImage(3,3) = constVal	! The value is 100 if (x,y)=(3,3), else 0.

    oldPixelCorners(:,:)%x = -99.0
    oldPixelCorners(:,:)%y = -99.0
    ! Transform of C1(x) -> D1(x-2)
    !   i.e., x_D1 = x_C1 - 2   (e.g., x_C1=2 => x_D1=0)
    !         y_D1 = y_C1
    mtrxLinTrans = mtrxNormal
    mtrxLinTrans(1,1) = 4.0
    vecTranslate = 0.0
    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldImageEdgesInfo, newImageEdgesInfo, mtrxLinTrans, vecTranslate)

    newImage(:,:) = 0.0

    call regridCartesian(oldPixelCorners, oldImage, newImage, status&
       , testFlagArg, trim(averagingStyle))

    if ( status .ne. 0 ) then
      print *, 'status=',status
      call error("err"//trim(errHead), '')
    end if

    oldStatInfo = getAryStatInfo(oldImage)	! defined in array_utils
    newStatInfo = getAryStatInfo(newImage)	! defined in array_utils

    if (.not. isNearlyEqual(oldStatInfo%realsum, newStatInfo%realsum / 4.0, maxPrecision)) then	! Increases by a factor of 4 in averagingStyle='rate'.
    ! if ( oldStatInfo%realsum .ne. newStatInfo%realsum / 4.0) then	! Increases by a factor of 4 in averagingStyle='rate'.
      print *, 'oldStatInfo%realsum, newStatInfo%realsum = ' &
              , oldStatInfo%realsum, newStatInfo%realsum
! print *,'oldImage=',oldImage 
! print *,'newImage=',newImage
      do i=9,15
        print *,'i=',i,'newImage(i,1:10)=',newImage(i,1:10)
      end do
      call error("err"//trim(errHead), '')
    end if
  
    oldIndices(:) = oldStatInfo%maxindices
    newIndices(:) = newStatInfo%maxindices
    a = real(newIndices(1),double)/4.0/oldIndices(1)	! Ideally, 12 and 3.
    if ( oldIndices(2) .ne. newIndices(2)  .or. &
         (a < 0.6) .or. (a > 1.4) ) then
      print *,'oldIndices=',oldIndices(:)
      print *,'newIndices=',newIndices(:)
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "RegridCartesian022a"
    print *,"=== "//trim(errHead)//": Scaling along +X by 4. (averagingStyle=count)"

    newImage(:,:) = 0.0

    call regridCartesian(oldPixelCorners, oldImage, newImage, status&
       , testFlagArg, averagingStyle='count')

    if ( status .ne. 0 ) then
      print *, 'status=',status
      call error("err"//trim(errHead), '')
    end if

    oldStatInfo = getAryStatInfo(oldImage)	! defined in array_utils
    newStatInfo = getAryStatInfo(newImage)	! defined in array_utils

    if (.not. isNearlyEqual(oldStatInfo%realsum, newStatInfo%realsum, maxPrecision)) then	! The total conts are preserved in averagingStyle='count'.
      print *, 'oldStatInfo%realsum, newStatInfo%realsum = ' &
              , oldStatInfo%realsum, newStatInfo%realsum
! print *,'oldImage=',oldImage 
! print *,'newImage=',newImage
      do i=9,15
        print *,'i=',i,'newImage(i,1:10)=',newImage(i,1:10)
      end do
      call error("err"//trim(errHead), '')
    end if
  
    oldIndices(:) = oldStatInfo%maxindices
    newIndices(:) = newStatInfo%maxindices
    a = real(newIndices(1),double)/4.0/oldIndices(1)	! Ideally, 12 and 3.
    if ( oldIndices(2) .ne. newIndices(2)  .or. &
         (a < 0.6) .or. (a > 1.4) ) then
      print *,'oldIndices=',oldIndices(:)
      print *,'newIndices=',newIndices(:)
      call error("err"//trim(errHead), '')
    end if


    !----------------------------------------------------------
    errHead = "RegridCartesian031a"
    print *,"=== "//trim(errHead)//": Rotation by 45deg anti-clockwise (or clockwise of the coordinate)."

    deallocate(oldImage, newImage, oldPixelCorners)
    iSize = 9
    !allocate(oldImage(iSize,iSize), newImage(iSize*2-1,iSize*2-1))	! The index starts at 1 (NOT -4 etc.).
    allocate(oldImage(iSize,iSize), newImage(17,17))	! The index starts at 1 (NOT -4 etc.).
    allocate(oldPixelCorners(iSize+1, iSize+1))	! nb.: the size is larger by 1.
    lEdge = -4.5
    uEdge =  4.5
    oldImageEdgesInfo = getAryEdgesInfo(oldImage, lEdge=lEdge, uEdge=uEdge)
    lEdge = -8.5*2
    uEdge =  8.5*2
    newImageEdgesInfo = getAryEdgesInfo(newImage, lEdge=lEdge, uEdge=uEdge)

    constVal = 100.0
    oldImage(:,:) = 0.0
    do i=1, iSize
      oldImage(i,i) = constVal	! The value is 100 if (x,y)=(A,A), else 0.
    end do

    oldPixelCorners(:,:)%x = -99.0
    oldPixelCorners(:,:)%y = -99.0
    ! Transform of C1(x) -> D1(x-2)
    !   i.e., x_D1 = x_C1 - 2   (e.g., x_C1=2 => x_D1=0)
    !         y_D1 = y_C1
    mtrxLinTrans = mtrxNormal
    a = sqrt(2.0d0)/2	! 45 degree rotation.
    mtrxLinTrans(1,:) =  (/  a, a /)
    mtrxLinTrans(2,:) =  (/ -a, a /)
    vecTranslate = 0.0
    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldImageEdgesInfo, newImageEdgesInfo, mtrxLinTrans, vecTranslate)

    newImage(:,:) = 0.0

    call regridCartesian(oldPixelCorners, oldImage, newImage, status&
       , testFlagArg, 'count')
       !, testFlagArg, trim(averagingStyle))	! With 'rate' (here), it will underestimate the counts by a factor of 4.

    if ( status .ne. 0 ) then
      print *, 'status=',status
      call error("err"//trim(errHead), '')
    end if

    oldStatInfo = getAryStatInfo(oldImage)	! defined in array_utils
    newStatInfo = getAryStatInfo(newImage)	! defined in array_utils

    if (.not. isNearlyEqual(oldStatInfo%realsum, newStatInfo%realsum, maxPrecision)) then
      print *, 'oldStatInfo%realsum, newStatInfo%realsum = ' &
              , oldStatInfo%realsum, newStatInfo%realsum
! print *,'oldImage=',oldImage 
! print *,'newImage=',newImage

      do i=1,iSize
        print *,'i=',i,'oldImage(i,:)=',oldImage(i,:)
      end do
      do i=1,size(newImage,2)
        print *,'i=',i,'newImage(i,8:11)=',newImage(i,8:11)
      end do

      call error("err"//trim(errHead), '')
    end if

    a = newStatInfo%maxval / constVal / sqrt(2.0)	! The coordinates are practically condensed by a factor of sqrt(2), so the new maximum value must be larger by a factor of sqrt(2) very roughly.  In practice in computation it is more steppy, so a reasonable amount of allowance should be allowed.
    if ( ( a < 0.75 ) .or. ( a > 1.25 ) ) then
      print *, 'newStatInfo%maxval = ', newStatInfo%maxval
      print *, 'newStatInfo%maxindices = ', newStatInfo%maxindices

      do i=1,iSize
        print *,'i=',i,'oldImage(i,:)=',oldImage(i,:)
      end do
      do i=1,size(newImage,2)
        print *,'i=',i,'newImage(i,8:11)=',newImage(i,8:11)
      end do

      call error("err"//trim(errHead), '')
    end if

    newIndices(:) = newStatInfo%maxindices
    if ( (newIndices(1) < 8) .or. (newIndices(1) > 10)  .or. &
         (newIndices(2) .ne. 9) ) then
      print *, 'newStatInfo%maxindices = ', newStatInfo%maxindices

      do i=1,iSize
        print *,'i=',i,'oldImage(i,:)=',oldImage(i,:)
      end do
      do i=1,size(newImage,2)
        print *,'i=',i,'newImage(i,8:11)=',newImage(i,8:11)
      end do

      call error("err"//trim(errHead), '')
    end if
  
    !! for Debug...
    ! do j=8,10
    !   print *,'j=',j,'newImage(5:13,j)=',newImage(5:13,j)
    ! end do
    ! ==>
    !  j=  8 newImage(5:13,j)=   0.0000000   0.0000000   0.0000000   0.0000000   0.0000000   0.0000000   0.0000000   0.0000000   0.0000000
    !  j=  9 newImage(5:13,j)=   0.0000000  99.7474670   1.2893216E+02   1.6274167E+02   1.1715720E+02   1.6274174E+02   1.2893216E+02  99.7475128   0.0000000
    !  j= 10 newImage(5:13,j)=   0.0000000   0.0000000   0.0000000   0.0000000   0.0000000   0.0000000   0.0000000   0.0000000   0.0000000


    !----------------------------------------------------------
    errHead = "RegridCartesian041a"
    print *,"=== "//trim(errHead)//": Parallel move for X (x2=x1-2) onto the identical frames. (Case 1)"
    ! NOTE: Those pixel values, of which the corresponding place in the new frame is out of its boundary, are simply thrown away.

    deallocate(oldImage, newImage, oldPixelCorners)
    lboundOldImage = -4
    uboundOldImage = 4
    iSize = uboundOldImage - lboundOldImage + 1
    allocate(oldImage(lboundOldImage:uboundOldImage, lboundOldImage:uboundOldImage))
    allocate(newImage(lboundOldImage:uboundOldImage, lboundOldImage:uboundOldImage))
    allocate(oldPixelCorners(iSize+1, iSize+1))	! nb.: the size is larger by 1.
    oldImageEdgesInfo = getAryEdgesInfo(oldImage, lboundIndex=lbound(oldImage), style='wcs')
    newImageEdgesInfo = getAryEdgesInfo(newImage, lboundIndex=lbound(newImage), style='wcs')

    constVal = 7.0
    oldImage(:,:) = 0.0
    oldImage(:,1) = constVal	! The value is 7 if y=1, else 0.

    oldPixelCorners(:,:)%x = -99.0
    oldPixelCorners(:,:)%y = -99.0
    ! Transform of C1(x) -> D1(x-2)
    !   i.e., x_D1 = x_C1 - 2   (e.g., x_C1=2 => x_D1=0)
    !         y_D1 = y_C1
    mtrxLinTrans = mtrxNormal
    vecTranslate = (/ -2, 0 /)
    call calcOldPixelCornersAffine(oldPixelCorners &
       , oldImageEdgesInfo, newImageEdgesInfo, mtrxLinTrans, vecTranslate)

    newImage(:,:) = 0.0

    call regridCartesian(oldPixelCorners, oldImage, newImage, status&
       , testFlagArg, trim(averagingStyle))

    if ( status .ne. 0 ) then
      print *, 'status=',status
      call error("err"//trim(errHead), '')
    end if

    oldStatInfo = getAryStatInfo(oldImage)	! defined in array_utils
    newStatInfo = getAryStatInfo(newImage)	! defined in array_utils

    if (.not. isNearlyEqual(oldStatInfo%realsum, newStatInfo%realsum + constVal*2, maxPrecision)) then	! 2 pixels are lost in transform.
    ! if ( oldStatInfo%realsum .ne. newStatInfo%realsum + constVal*2) then	! 2 pixels are lost in transform.
    !if ( oldStatInfo%realsum .ne. newStatInfo%realsum ) then
      print *, 'oldStatInfo%realsum, newStatInfo%realsum = ' &
              , oldStatInfo%realsum, newStatInfo%realsum &
              , oldStatInfo%realsum - (newStatInfo%realsum + constVal*2)
      call error("err"//trim(errHead), '')
    end if

    if ( any(newImage(-4:2,1) .ne. constVal) ) then
      print *, 'All the points should be (ideally) ', constVal
      print *, 'However it may not be quite achieved in 64-bit machines.'
      print *, 'newImage(-4:2,1) = ', newImage(-4:2,1)
      call warning("warning"//trim(errHead), '')
    end if
    if ( any(newImage(3:,1) .ne. 0.0) ) then
      print *, 'newImage(3:,1) = ', newImage(3:,1)
      call error("err"//trim(errHead), '')
    end if

    if ( any(newImage(:,-1) .ne. 0.0) ) then
      print *, 'newImage(:,-1) = ',  newImage(:,-1)
      call error("err"//trim(errHead), '')
    end if

  end subroutine test_regridCartesian


end program testregridders1
